# -*- coding: utf-8 -*-
# (c) 2017,2018 Andreas Motl <andreas@hiveeyes.org>
# (c) 2017,2018 Richard Pobering <richard@hiveeyes.org>
# License: GNU Affero General Public License, Version 3
import os
import json
import logging
from munch import Munch
from pprint import pformat
import paho.mqtt.client as mqtt

try:
    from urlparse import urlsplit
except ImportError:
    from urllib.parse import urlsplit

log = logging.getLogger(__name__)


class MQTTAdapter(object):

    capabilities = ['readings']

    def __init__(self, uri, keepalive=60, client_id_prefix=None, dry_run=False):

        address = urlsplit(uri)
        self.host       = address.hostname
        self.port       = address.port or 1883
        self.username   = address.username
        self.password   = address.password
        self.topic      = address.path

        self.keepalive  = keepalive

        self.client_id_prefix = client_id_prefix or 'luftdatenpumpe'

        self.dry_run = dry_run

        self.connect()

    def emit(self, reading):
        """
        Will publish these kind of messages to the MQTT bus:
        {
            "time": "2018-12-03T01:49:00Z",
            "location_id": 9564,
            "sensor_id": 18870,
            "sensor_type": "DHT22",
            "geohash": "u3qcs53rp",
            "altitude": 85.0,
            "temperature": 2.8,
            "humidity": 90.5
        }
        """

        # Build MQTT message.
        message = Munch()

        # Station info
        for key, value in reading.station.items():
            if isinstance(value, dict):
                message.update(value)
            else:
                message[key] = value

        message['location_id'] = message['station_id']
        del message['station_id']

        # Measurement fields
        message.update(reading.data)

        # Publish to MQTT bus.
        if self.dry_run:
            log.info('Dry-run. Would publish record:\n{}'.format(pformat(message)))
        else:
            # FIXME: Don't only use ``sort_keys``. Also honor the field names of the actual readings by
            # putting them first. This is:
            # - "P1" and "P2" for "sensor_type": "SDS011"
            # - "temperature" and "humidity" for "sensor_type": "DHT22"
            # - "temperature", "humidity", "pressure" and "pressure_at_sealevel" for "sensor_type": "BME280"
            mqtt_message = json.dumps(message)
            self.publish(mqtt_message)

    def flush(self, final=False):
        pass

    def connect(self):

        if self.dry_run:
            return

        # Create a mqtt client object
        pid = os.getpid()
        client_id = '{}:{}'.format(self.client_id_prefix, str(pid))
        self.mqttc = mqtt.Client(client_id=client_id, clean_session=True)

        # Handle authentication
        if self.username:
            self.mqttc.username_pw_set(self.username, self.password)

        # Connect to broker
        self.mqttc.connect(self.host, self.port, self.keepalive)

        self.mqttc.loop_start()

    def close(self):
        self.mqttc.disconnect()

    def effective_topic(self, topic=None):
        parts = []
        base_topic = self.topic.strip(u'/')
        if base_topic:
            parts.append(base_topic)
        if topic:
            parts.append(topic)
        return u'/'.join(parts)

    def publish(self, message, topic=None):
        topic = self.effective_topic(topic)
        log.debug('Publishing to topic "{}": {}'.format(topic, message))
        return self.mqttc.publish(topic, message)

    def subscribe(self, topic=None):
        topic = self.effective_topic(topic)
        log.info('Subscribing to topic "{}"'.format(topic))
        return self.mqttc.subscribe(topic)



# -*- coding: utf-8 -*-
# (c) 2019 Andreas Motl <andreas@hiveeyes.org>
# (c) 2019 Richard Pobering <richard@hiveeyes.org>
# (c) 2019 Matthias Mehldau <wetter@hiveeyes.org>
# License: GNU Affero General Public License, Version 3
import logging
from copy import deepcopy
from datetime import datetime, timedelta
from operator import itemgetter

from requests import HTTPError
from rfc3339 import rfc3339
from munch import munchify, Munch
from tablib import Dataset
from urllib.parse import urljoin
from collections import OrderedDict

from luftdatenpumpe.source.common import AbstractLuftdatenPumpe
from luftdatenpumpe.util import slugify, grouper, chunks

log = logging.getLogger(__name__)


class EEAAirQualityPumpe(AbstractLuftdatenPumpe):
    """
    Ingest air quality measurements from the
    European Environment Agency (EEA).

    - http://discomap.eea.europa.eu/map/fme/AirQualityExport.htm
    - http://ftp.eea.europa.eu/www/aqereporting-3/AQeReporting_products_2018_v1.pdf
    - http://dd.eionet.europa.eu/vocabulary/aq/pollutant

    """

    # Sensor network identifier.
    network = 'eea'

    # Download service REST API URI
    uri = 'https://ereporting.blob.core.windows.net/downloadservice/'

    timeout = 60

    def get_index(self):
        return self.send_request()

    def send_request(self, endpoint=None, params=None):
        url = urljoin(self.uri, endpoint)
        log.info(f'Requesting EEA at {url}')
        params = params or {}

        response = self.session.get(url, params=params, timeout=self.timeout)
        if response.status_code != 200:
            try:
                reason = response.json()
            except:
                reason = 'unknown'
            message = f'Request failed: {reason}'
            log.error(message)
            response.raise_for_status()

        return response.text

    def get_stations(self):
        """
        https://ereporting.blob.core.windows.net/downloadservice/metadata.csv
        """

        payload = self.send_request('metadata.csv')

        # Read CSV file into tablib's Dataset and cast to dictionary representation.
        try:
            data = Dataset().load(payload, format='csv', delimiter=',')
        except Exception:
            log.exception(f'Error reading or decoding station CSV')
            return

        # Apply data filter.
        data = self.apply_filter(data)

        df_grouped = data.df.groupby(['Countrycode', 'AirQualityNetwork', 'AirQualityStation'], as_index=False)

        stations = []
        for name, group in self.wrap_progress(df_grouped):
            first_sensor = group.head(1)
            data = first_sensor.to_dict(orient='records')[0]
            station_info = munchify({
                'station_id': data.pop('AirQualityStation'),
                'station_label': data.pop('Namespace'),
                'network': data.pop('AirQualityNetwork'),
                'nat_code': data.pop('AirQualityStationNatCode'),
                'eoi_code': data.pop('AirQualityStationEoICode'),
                'position': {
                    'country': data.pop('Countrycode'),
                    'latitude': data.pop('Latitude'),
                    'longitude': data.pop('Longitude'),
                    'altitude': data.pop('Altitude'),
                    'area': data.pop('AirQualityStationArea'),
                    'projection': data.pop('Projection'),
                    'building_distance': data.pop('BuildingDistance'),
                    'kerb_distance': data.pop('KerbDistance'),
                }
            })

            # Transfer all other stuff.
            #for key, value in data.items():
            #    station_info[key] = value

            #print('station_info:', station_info)
            #continue

            self.enrich_station(station_info)

            # FIXME: Add sensors.
            #station_info.sensors = self.timeseries_to_sensors(upstream_station.properties.timeseries)

            #print(station_info)
            stations.append(station_info)

        # List of stations sorted by station identifier.
        stations = sorted(stations, key=itemgetter('station_id'))

        return stations

    def filter_rule(self, data):
        print('FILTER:', self.filter)
        df_filtered = data.df
        if self.filter and 'country' in self.filter:
            df_filtered = df_filtered[df_filtered['Countrycode'].isin(self.filter.country)]

        data.df = df_filtered
        return data



# -*- coding: utf-8 -*-
# (c) 2017-2019 Andreas Motl <andreas@hiveeyes.org>
# License: GNU Affero General Public License, Version 3
import json
import logging
import dataset
from copy import deepcopy
from munch import munchify
from collections import OrderedDict

from geoalchemy2 import Geometry    # Pulls PostGIS capabilities into SQLAlchemy. Don't remove.
from sqlalchemy_utils import drop_database

from luftdatenpumpe.util import sanitize_dbsymbol


log = logging.getLogger(__name__)


class DatabaseError(Exception):
    pass


class RDBMSStorage:

    capabilities = ['stations']

    def __init__(self, uri, network=None, dry_run=False):

        self.dry_run = dry_run

        self.network = network or 'default'
        self.realm = sanitize_dbsymbol(self.network)

        # Connect to database.
        self.db = dataset.connect(uri)

        self.artefacts = munchify([
            {'name': f'{self.realm}_stations', 'type': 'table', 'primary_id': 'station_id', 'primary_type': self.db.types.bigint},
            {'name': f'{self.realm}_osmdata', 'type': 'table', 'primary_id': 'station_id', 'primary_type': self.db.types.bigint},
            {'name': f'{self.realm}_sensors', 'type': 'table', 'primary_id': 'sensor_id', 'primary_type': self.db.types.bigint},
            {'name': f'{self.realm}_network', 'type': 'view'},
        ])

        # Ping database.
        self.ping_database()

        # Check for PostGIS extension.
        self.ensure_postgis()

        # Deploy the database schema.
        self.ensure_schema()

        self.tame_library_loggers()

        # Assign table handles.
        self.stationtable = self.db[f'{self.realm}_stations']
        self.sensorstable = self.db[f'{self.realm}_sensors']
        self.osmtable = self.db[f'{self.realm}_osmdata']

    def emit(self, station):
        return self.store_station(station)

    def flush(self, final=False):
        pass

    def ping_database(self):
        failure = None
        try:
            self.db.query('SELECT TRUE;')

        except Exception as ex:
            failure = f'Database operation failed. Reason: {ex}'
            log.exception(failure, exc_info=False)

        if failure:
            raise DatabaseError(failure)

    def ensure_postgis(self):

        # TODO: Add PostGIS extension. However, this requires superuser permissions.
        # self.db.query('CREATE EXTENSION IF NOT EXISTS postgis')

        failure = None
        try:
            result = self.db.query('SELECT PostGIS_Full_Version();')
            postgis_full_version = list(result)[0]['postgis_full_version']
            log.info(f'PostGIS version: {postgis_full_version}')

        except Exception as ex:

            message = 'Database operation failed'
            if 'function postgis_full_version() does not exist' in str(ex):
                message = 'PostGIS extension not installed'

            failure = f'{message}. Reason: {ex}'
            log.exception(failure, exc_info=False)

        if failure:
            raise DatabaseError(failure)

    def ensure_schema(self):

        # Create tables.
        for tableinfo in self.tableinfos:
            self.db.create_table(tableinfo.name, primary_id=tableinfo.primary_id, primary_type=tableinfo.get('primary_type'))

        # FIXME: Funny enough, this workaround helps after upgrading to PostgreSQL 11.
        repr(self.db._tables)

        # Enforce "xox_stations.(latitude,longitude,altitude)"  to be of "float" type.
        stations_table = self.db.get_table(f'{self.realm}_stations')
        stations_table.create_column('latitude', self.db.types.float)
        stations_table.create_column('longitude', self.db.types.float)
        stations_table.create_column('altitude', self.db.types.float)

        # Enforce "xox_sensors.(sensor_first_date,sensor_last_date)"  to be of "datetime" type.
        stations_table = self.db.get_table(f'{self.realm}_sensors')
        stations_table.create_column('sensor_first_date', self.db.types.datetime)
        stations_table.create_column('sensor_last_date', self.db.types.datetime)

        # Enforce "xox_osmdata.osm_id" and "xox_osmdata.osm_place_id" to be of "bigint" type.
        # Otherwise: ``psycopg2.DataError: integer out of range`` with 'osm_id': 2678458514
        osmdata_table = self.db.get_table(f'{self.realm}_osmdata')
        osmdata_table.create_column('osm_id', self.db.types.bigint)
        osmdata_table.create_column('osm_place_id', self.db.types.bigint)

        # Manually add "geopoint" field to the table because the "dataset"
        # package does not implement appropriate heuristics for that.

        # EPSG:4326
        # geography(Point,4326)
        self.db.query(f'ALTER TABLE {self.realm}_stations ADD COLUMN IF NOT EXISTS "geopoint" geography(Point,4326)')

        # TODO: Convert to EPSG:4258.

        # TODO: Convert to EPSG:31370.
        #self.db.query(f'ALTER TABLE {self.realm}_stations ADD COLUMN IF NOT EXISTS "geopoint_31370" geometry(Point,31370)')


    @property
    def tableinfos(self):
        for artefact in self.artefacts:
            if artefact.type == 'table':
                yield artefact

    @property
    def viewinfos(self):
        for artefact in self.artefacts:
            if artefact.type == 'view':
                yield artefact

    def store_station(self, station):

        # Make up geoposition on the fly
        # In mapping frameworks spatial coordinates are often in order of latitude and longitude.
        # In spatial databases spatial coordinates are in x = longitude, and y = latitude.
        # -- https://postgis.net/2013/08/18/tip_lon_lat/
        # Example: Stuttgart == POINT(9.17721 48.77928)
        if station.position and station.position.latitude and station.position.longitude:
            station.position.geopoint = 'POINT({} {})'.format(station.position.longitude, station.position.latitude)

            # TODO: Convert to EPSG:31370.
            #station.position.geopoint_31370 = f"ST_GeomFromText('POINT({station.position.longitude} {station.position.latitude})', 4326)"
            #station.position.geopoint_31370 = 'POINT({} {})'.format(station.position.longitude, station.position.latitude)

        # Debugging
        #log.info('station: %s', station)

        # Station table
        stationdata = OrderedDict()
        stationdata['station_id'] = station.station_id
        for key, value in station.items():
            if key.startswith('name'):
                stationdata[key] = value
        stationdata.update(station.position)
        self.stationtable.upsert(stationdata, ['station_id'])

        # Sensors table
        for sensor in station['sensors']:
            sensordata = OrderedDict()
            sensordata['station_id'] = station.station_id
            sensordata.update(sensor)
            self.sensorstable.upsert(sensordata, ['sensor_id'])

        # OSM table
        if 'location' in station:
            osmdata = OrderedDict()
            osmdata['station_id'] = station.station_id
            location = deepcopy(station.location)
            for key, value in location.address.items():
                #key = 'address_' + key
                location[key] = value
            del location['address']

            if 'address_more' in location:
                osmdata['address_more'] = json.dumps(location.address_more)
                del location['address_more']

            # TODO: Also store bounding box
            del location['boundingbox']

            osmdata.update(location)

            # Prefix designated column name with 'osm_'.
            osmdata_real = OrderedDict()
            for key, value in osmdata.items():
                if key not in ['station_id']:
                    if not key.startswith('osm_'):
                        key = 'osm_' + key
                osmdata_real[key] = value

            self.osmtable.upsert(osmdata_real, ['station_id'])

    def dump_tables(self):
        for table in [self.stationtable, self.sensorstable, self.osmtable]:
            print('### Table name:', table.name)
            for record in table.all():
                print(self.json_markdown_dumps(record))
                print()

    def demo_query(self):

        #print(dir(self.stationtable.table))

        # https://dataset.readthedocs.io/en/latest/quickstart.html#running-custom-sql-queries
        expression = f'SELECT * FROM {self.realm}_stations, {self.realm}_osmdata WHERE {self.realm}_stations.station_id = {self.realm}_osmdata.station_id'
        print('### Expression:', expression)
        result = self.db.query(expression)
        for record in result:
            print(self.json_markdown_dumps(record))

    @staticmethod
    def json_markdown_dumps(thing):
        tpl = "```\n{}\n```"
        content = json.dumps(thing, indent=4)
        return tpl.format(content)

    @classmethod
    def get_dialects(cls):
        results = []
        try:
            from sqlalchemy.dialects import __all__ as sql_dialects
            results = list(sql_dialects)
        except:
            pass
        #log.info('SQLAlchemy dialects: %s', results)
        return results

    def render_fields(self, data, carry_on=True):
        if not data:
            return ''
        payload = ', '.join(data)
        if carry_on:
            payload += ','
        return payload

    def create_views(self):

        prefix = self.realm

        # Collect fields from osmdata table.
        osmdata = self.db.query(f"""
            SELECT concat(TABLE_NAME, '.', COLUMN_NAME) AS name
            FROM INFORMATION_SCHEMA.COLUMNS
            WHERE TABLE_NAME='{prefix}_osmdata' AND COLUMN_NAME NOT IN ('station_id')
            ORDER BY ORDINAL_POSITION
        """)
        osmdata_columns = []
        for record in osmdata:
            osmdata_columns.append(record['name'])

        # Collect conditional aliases.
        conditional_fields_stage1 = []
        conditional_fields_stage2 = []
        if self.realm == 'irceline':
            conditional_fields_stage1.append(f'{prefix}_sensors.sensor_feature_label')
            conditional_fields_stage2.append(f"concat(sensor_feature_label, ' ', station_id_suffix) AS sos_feature_and_id")

        # Create unified view.
        view = f"""
        DROP VIEW IF EXISTS {prefix}_network;
        CREATE VIEW {prefix}_network AS

            WITH stage1 AS (
            SELECT

              -- Baseline fields.
              {prefix}_stations.station_id,
              {prefix}_stations.name, {prefix}_stations.country, 
              {prefix}_stations.longitude, {prefix}_stations.latitude, {prefix}_stations.altitude, 
              {prefix}_stations.geohash, {prefix}_stations.geopoint,
              {prefix}_sensors.sensor_id, {prefix}_sensors.sensor_type_id, {prefix}_sensors.sensor_type_name,
              {prefix}_sensors.sensor_first_date, {prefix}_sensors.sensor_last_date,
              {self.render_fields(conditional_fields_stage1)}

              -- Synthesized fields.
              concat('(#', CAST({prefix}_stations.station_id AS text), ')') AS station_id_suffix,
              concat('(', {prefix}_osmdata.osm_country_code, ')') AS country_code_suffix,

              -- OSM fields.
              {self.render_fields(osmdata_columns, carry_on=False)}

            FROM
              {prefix}_stations, {prefix}_osmdata, {prefix}_sensors

            WHERE
              {prefix}_stations.station_id = {prefix}_osmdata.station_id AND
              {prefix}_stations.station_id = {prefix}_sensors.station_id

            )

            SELECT
              *,
              concat(name, ' ', station_id_suffix) AS name_and_id,
              concat(osm_state, ' » ', osm_city) AS state_and_city,
              concat(osm_road, ', ', osm_city, ' ', station_id_suffix) AS road_and_name_and_id,
              concat(osm_country, ' (', osm_country_code, ')') AS country_and_countrycode,
              concat(concat_ws(', ', osm_state, osm_country), ' ', country_code_suffix) AS state_and_country,
              concat(concat_ws(', ', osm_city, osm_state, osm_country), ' ', country_code_suffix) AS city_and_state_and_country,
              {self.render_fields(conditional_fields_stage2)}
              ABS(DATE_PART('day', sensor_last_date - now())) <= 7 AS is_active

            FROM
              stage1

            ORDER BY
              osm_country_code, state_and_city, name_and_id, sensor_type_name

        """

        self.db.query(view)

    def grant_read_privileges(self, user):
        sql = """
        GRANT USAGE, SELECT ON ALL SEQUENCES IN SCHEMA public TO {user};
        GRANT SELECT ON ALL TABLES IN SCHEMA public TO {user};
        """.format(user=user)
        self.db.query(sql)

    def drop_tables(self):
        log.info('Dropping views and tables in database %s', self.db)

        for viewinfo in self.viewinfos:
            self.db.query('DROP VIEW IF EXISTS {}'.format(viewinfo.name))

        for tableinfo in self.tableinfos:
            table = self.db.get_table(tableinfo.name)
            table.drop()

    def drop_database(self):
        drop_database(self.db.engine.url)

    def drop_data(self):
        """
        Prune data from all tables.
        """
        for tablename in self.db.tables:
            if tablename.startswith(f'{self.realm}_'):
                log.info('Deleting data from table {}'.format(tablename))
                self.db.query('DELETE FROM {}'.format(tablename))

    def get_all_view_names(self):
        from sqlalchemy import inspect
        inspector = inspect(self.db.engine)
        return inspector.get_view_names()

    def get_database_name(self):
        raise NotImplementedError('Defunct.')

        # https://stackoverflow.com/questions/53554458/sqlalchemy-get-database-name-from-engine/53558430#53558430
        from sqlalchemy import inspect
        inspector = inspect(self.db.engine)
        print(inspector)
        print(dir(inspector))
        print('url:', self.db.engine.url)
        #name = inspector.default_schema_name
        name = inspector.name
        return name

    def get_all_table_names(self):
        raise NotImplementedError('Defunct.')

        for table in self.db._tables.values():
            log.info('Dropping table %s', table.name)
            try:
                table.drop()
            except:
                log.warning('Skipped dropping table/view "%s"', table.name)

    @staticmethod
    def tame_library_loggers():
        logger = logging.getLogger('alembic.runtime.migration')
        logger.disabled = True



# -*- coding: utf-8 -*-
# (c) 2018-2019 Andreas Motl <andreas@hiveeyes.org>
# License: GNU Affero General Public License, Version 3
import json
import logging
from urllib.parse import urlparse

from tqdm import tqdm
from munch import Munch

from luftdatenpumpe.target.influxdb import InfluxDBStorage
from luftdatenpumpe.target.json import json_formatter, json_grafana_formatter_vt, json_grafana_formatter_kn, JsonFlexFormatter
from luftdatenpumpe.target.mqtt import MQTTAdapter
from luftdatenpumpe.target.rdbms import RDBMSStorage
from luftdatenpumpe.target.stream import StreamTarget

log = logging.getLogger(__name__)


class LuftdatenEngine:

    def __init__(self, network, domain, targets, fieldmap=None, batch_size=None, progressbar=False, dry_run=False):
        self.network = network
        self.domain = domain
        self.targets = targets
        self.fieldmap = fieldmap
        self.batch_size = batch_size or 1
        self.progressbar = progressbar
        self.dry_run = dry_run

    def process(self, data):

        # Configure target subsystems.
        targets = []
        for target_expression in self.targets:
            log.info(f'Configuring data sink "{target_expression}" with domain "{self.domain}"')

            try:
                target = self.resolve_target_handler(target_expression, dry_run=self.dry_run)

            except Exception as ex:
                log.exception(f'Resolving data sink "{target_expression}" failed. Reason: {ex}', exc_info=False)
                continue

            if target is None:
                log.error(f'Could not resolve data sink {target_expression}')
                continue

            if self.domain not in target.capabilities:
                log.warning(f'Data sink {target_expression} does not handle domain "{self.domain}"')
                continue

            targets.append(target)

        # Sanity checks.
        if not targets:
            raise ValueError('No data sinks enabled, please check the log file for errors')

        # Emit to active target subsystems.
        log.info('Emitting to target data sinks, this might take some time')

        if self.progressbar:
            data = tqdm(list(data))

        item_count = 0
        for item in data:

            for target in targets:
                target.emit(item)
            item_count += 1

            # Preliminary flush each $batch_size items.
            if item_count % self.batch_size == 0:
                for target in targets:
                    target.flush()

        # Signal readiness to each target subsystem.
        for target in targets:
            target.flush(final=True)

        log.info('Processed {} records'.format(item_count))

    def resolve_target_handler(self, target, dry_run=False):
        handler = None

        url = Munch(urlparse(target)._asdict())
        log.debug('Resolving target: %s', json.dumps(url))

        formatter = json_formatter
        if '+' in url.scheme:
            format, scheme = url.scheme.split('+')
            url.scheme = scheme

            if format.startswith('json.grafana.vt'):
                formatter = json_grafana_formatter_vt
            elif format.startswith('json.grafana.kn'):
                formatter = json_grafana_formatter_kn
            elif format.startswith('json.flex'):
                formatter = JsonFlexFormatter(self.fieldmap).formatter
            elif format.startswith('json'):
                formatter = json_formatter

            # Do we need this anymore?
            #formatter.format = format

        #effective_url = urlunparse(url.values())

        if url.scheme == 'stream':

            # FIXME: There might be dragons?
            import sys
            stream = eval(url.netloc)

            handler = StreamTarget(stream, formatter)

        elif url.scheme in RDBMSStorage.get_dialects():
            handler = RDBMSStorage(target, network=self.network, dry_run=dry_run)

        elif url.scheme == 'mqtt':
            handler = MQTTAdapter(target, dry_run=dry_run)

        elif url.scheme == 'influxdb':
            handler = InfluxDBStorage(target, network=self.network, dry_run=dry_run)

        return handler



# -*- coding: utf-8 -*-
# (c) 2018-2019 Andreas Motl <andreas@hiveeyes.org>
# License: GNU Affero General Public License, Version 3
import json
import types
import logging
from collections import OrderedDict

log = logging.getLogger(__name__)


def json_formatter(data):
    if isinstance(data, types.GeneratorType):
        data = list(data)
    return json.dumps(data, indent=2)


def curated_station_list(stations):
    seen = {}
    for station in stations:
        if station.station_id in seen:
            continue
        if 'name' not in station:
            station.name = u'Station #{}, {}'.format(station.station_id, station.position.country)
        seen[station.station_id] = True
        yield station


class JsonFlexFormatter:

    def __init__(self, fieldmap):
        self.fieldmap = fieldmap

    def formatter(self, stations):
        entries = []
        for station in curated_station_list(stations):
            entry = OrderedDict()
            for key_left, key_right in self.fieldmap.items():

                # Compute value, optionally applying conversion.
                conversion = None
                if '|' in key_right:
                    key_right, conversion = key_right.split('|')
                value = station[str(key_right)]
                if conversion is not None:
                    converter = eval(conversion)
                    value = converter(value)

                entry[str(key_left)] = value

            entries.append(entry)
        return json_formatter(entries)


def json_grafana_formatter_vt(stations):
    """
    Format list of stations in JSON format made of value/text items,
    suitable for use as a Grafana JSON data source.

    Example::

        luftdatenpumpe stations --source=postgresql://luftdatenpumpe@localhost/weatherbase --target=json.grafana.vt+stream://sys.stdout

    Emits items like::

        {
            "value": 22,
            "text": "Steglen, Haslach, Herrenberg, Vereinbarte  der Stadt Herrenberg, Baden-W\u00fcrttemberg, DE"
        }

    """
    entries = []
    for station in curated_station_list(stations):
        entry = {'value': str(station.station_id), 'text': station.name}
        entries.append(entry)
    return json_formatter(entries)


def json_grafana_formatter_kn(stations):
    """
    Format list of stations in JSON format made of key/name items,
    suitable for use as a mapping in Grafana Worldmap Panel.

    Example::

        luftdatenpumpe stations --source=postgresql://luftdatenpumpe@localhost/weatherbase --target=json.grafana.kn+stream://sys.stdout

    Emits items like::

        {
            "key": "22",
            "name": "Steglen, Haslach, Herrenberg, Vereinbarte  der Stadt Herrenberg, Baden-W\u00fcrttemberg, DE"
        }

    """
    entries = []
    for station in curated_station_list(stations):
        entry = {'key': str(station.station_id), 'name': station.name}
        entries.append(entry)
    return json_formatter(entries)



# -*- coding: utf-8 -*-
# (c) 2017-2019 Andreas Motl <andreas@hiveeyes.org>
# (c) 2017-2019 Richard Pobering <richard@hiveeyes.org>
# License: GNU Affero General Public License, Version 3
from munch import munchify

from luftdatenpumpe.target.rdbms import RDBMSStorage
from luftdatenpumpe.util import sanitize_dbsymbol


def stations_from_rdbms(dsuri, prefix):
    """
    Re-make station information from PostgreSQL database.

    Example::

        {
            "station_id": 28,
            "name": "Ulmer Stra\u00dfe, Wangen, Stuttgart, Baden-W\u00fcrttemberg, DE",
            "position": {
                "latitude": 48.778,
                "longitude": 9.236,
                "altitude": 223.7,
                "country": "DE",
                "geohash": "u0wt6pv2qqhz"
            },
            "location": {
                "place_id": "10504194",
                "licence": "Data \u00a9 OpenStreetMap contributors, ODbL 1.0. https://osm.org/copyright",
                "osm_type": "way",
                "osm_id": "115374184",
                "lat": "48.777845",
                "lon": "9.23582396156841",
                "display_name": "Kulturhaus ARENA, 241, Ulmer Stra\u00dfe, Wangen, Stuttgart, Regierungsbezirk Stuttgart, Baden-W\u00fcrttemberg, 70327, Deutschland",
                "boundingbox": [
                    "48.7775199",
                    "48.778185",
                    "9.2353783",
                    "9.236272"
                ],
                "address": {
                    "country_code": "DE",
                    "country": "Deutschland",
                    "state": "Baden-W\u00fcrttemberg",
                    "state_district": "Regierungsbezirk Stuttgart",
                    "county": "Stuttgart",
                    "postcode": "70327",
                    "city": "Stuttgart",
                    "city_district": "Wangen",
                    "suburb": "Wangen",
                    "road": "Ulmer Stra\u00dfe",
                    "house_number": "241",
                    "neighbourhood": "Wangen"
                },
                "address_more": {
                    "building": "Kulturhaus ARENA"
                }
            },
            "sensors": [
                {
                    "sensor_id": 658,
                    "sensor_type": "SDS011"
                },
                {
                    "sensor_id": 657,
                    "sensor_type": "DHT22"
                }
            ]
        }
    """

    for station in fetch_from_rdbms(dsuri, prefix):
        entry = {
            "station_id": station.station_id,
            "name": station.name,
            "position": {
                "latitude": station.latitude,
                "longitude": station.longitude,
                "altitude": station.altitude,
                "country": station.country,
                "geohash": station.geohash
            }
        }
        yield munchify(entry)


def stations_from_rdbms_flex(dsuri, prefix):
    return fetch_from_rdbms(dsuri, prefix)


def fetch_from_rdbms(dsuri, prefix):
    # Sanitize table name.
    prefix = sanitize_dbsymbol(prefix)

    sql = f"""
        SELECT *
        FROM {prefix}_network
        ORDER BY
          {prefix}_network.station_id
    """

    storage = RDBMSStorage(dsuri)
    for station in storage.db.query(sql):
        yield munchify(station)



# -*- coding: utf-8 -*-
# (c) 2017-2019 Andreas Motl <andreas@hiveeyes.org>
# (c) 2017-2019 Richard Pobering <richard@hiveeyes.org>
# License: GNU Affero General Public License, Version 3
import sys
import logging
from urllib.parse import urlparse

import redis
from tqdm import tqdm
from requests_cache import CachedSession
from luftdatenpumpe.geo import geohash_encode, resolve_location, improve_location, format_address

from luftdatenpumpe import __appname__ as APP_NAME
from luftdatenpumpe import __version__ as APP_VERSION


log = logging.getLogger(__name__)


class AbstractLuftdatenPumpe:

    network = None
    uri = None

    def __init__(self, source=None, filter=None, reverse_geocode=False, progressbar=False, quick_mode=False, dry_run=False):
        self.source = source
        self.reverse_geocode = reverse_geocode
        self.dry_run = dry_run
        self.progressbar = progressbar
        self.filter = filter

        # Quick mode only imports the first few datasets to speed things up.
        self.quick_mode = quick_mode

        # Cache responses from the luftdaten.info API for five minutes.
        # TODO: Make backend configurable.

        # Configure User-Agent string.
        user_agent = APP_NAME + '/' + APP_VERSION

        # Use hostname of url as cache prefix.
        cache_name = urlparse(self.uri).netloc

        # Configure cached requests session.
        self.session = CachedSession(
            cache_name=cache_name, backend='redis', expire_after=300,
            user_agent=user_agent)

        # Disable request cache by overriding it with a vanilla requests session.
        #import requests; self.session = requests.Session()

        # Gracefully probe Redis for availability if cache is enabled.
        if hasattr(self.session, 'cache'):
            try:
                self.session.cache.responses.get('test')
            except redis.exceptions.ConnectionError as ex:
                log.error('Unable to connect to Redis: %s', ex)
                sys.exit(2)

    def get_readings(self):

        if self.source == 'api':
            data = self.get_readings_from_api()

        elif self.source.startswith('file://'):
            path = self.source.replace('file://', '')
            data = self.get_readings_from_csv(path)

        else:
            raise ValueError('Unknown data source: {}'.format(self.source))

        if data is None:
            raise KeyError('No data selected. Please check connectivity and filter definition.')

        return data

    def get_readings_from_api(self):
        raise NotImplementedError(f'Readings not implemented by sensor network adapter "{self.network}".')

    def get_readings_from_csv(self):
        raise NotImplementedError(f'Readings not implemented by sensor network adapter "{self.network}".')

    def enrich_station(self, station):

        # Sanity checks.
        if ('latitude' not in station.position or 'longitude' not in station.position) or \
           (station.position.latitude is None or station.position.longitude is None):

            # Just emit this message once per station.
            StationGeocodingFailed.emit_warning(station.station_id)
            return

        # Compute geohash.
        station.position.geohash = geohash_encode(station.position.latitude, station.position.longitude)

        # Compute human readable location name.
        if self.reverse_geocode:

            try:

                # Reverse-geocode position.
                station.location = resolve_location(
                    latitude=station.position.latitude,
                    longitude=station.position.longitude,
                    geohash=station.position.geohash,
                    country_code=station.position.get('country'),
                )

                # Improve location information.
                improve_location(station.location)

                # Format address into single label.
                station.name = format_address(station.location)

                # Apply some fixups.
                try:
                    if station.name.lower() == station.position.country.lower():
                        del station['name']
                except:
                    pass

            except Exception:
                log.exception(f'Failed computing humanized name for station {station}')

            if 'name' not in station:
                station.name = f'Station #{station.station_id}'
                try:
                    station.name += ', ' + station.position.country
                except:
                    pass

    def apply_filter(self, data):
        if self.filter:
            return self.filter_rule(data)
        else:
            return data

    def wrap_progress(self, data, stepsize=False):
        # Optionally, add progress reporting.
        if self.progressbar:
            data = list(data)
            total = len(data)
            if stepsize is not False:
                total *= stepsize
            log.info(f'Processing {total} items')
            data = tqdm(data, unit_scale=stepsize)
        return data


class StationGeocodingFailed:

    stations_with_warnings = {}

    @classmethod
    def emit_warning(cls, station_id):
        if station_id in cls.stations_with_warnings:
            return
        log.warning(f'Incomplete station position, skipping geospatial enrichment for station "{station_id}"')
        cls.stations_with_warnings[station_id] = True



# -*- coding: utf-8 -*-
# (c) 2017-2019 Andreas Motl <andreas@hiveeyes.org>
# (c) 2017-2019 Richard Pobering <richard@hiveeyes.org>
# (c) 2019 Matthias Mehldau <wetter@hiveeyes.org>
# License: GNU Affero General Public License, Version 3
import re
import json
import logging
from tqdm import tqdm
from munch import Munch
from tablib import Dataset
from operator import itemgetter
from luftdatenpumpe.source.common import AbstractLuftdatenPumpe
from luftdatenpumpe.util import exception_traceback, find_files, is_nan

log = logging.getLogger(__name__)


class LuftdatenPumpe(AbstractLuftdatenPumpe):
    """
    Ingest air quality measurements from the
    luftdaten.info citizen network.

    - https://luftdaten.info/
    - https://community.hiveeyes.org/t/erneuerung-der-luftdatenpumpe/1199
    - https://community.hiveeyes.org/t/ldi-data-plane-v2/1412

    """

    # Sensor network identifier.
    network = 'ldi'

    # Live data API URI for luftdaten.info.
    uri = 'https://api.luftdaten.info/static/v1/data.json'

    def get_stations(self):

        stations = {}
        field_candidates = ['station_id', 'name', 'position', 'location']
        for reading in self.get_readings():

            # LDI Location ID is "station_id" here.
            station_id = reading.station.station_id

            # New station found: Acquire its information from the reading itself.
            if station_id not in stations:
                station = Munch()
                for field in field_candidates:
                    if field in reading.station:
                        station[field] = reading.station[field]
                station.sensors = []

                # Record station.
                stations[station_id] = station

            # Use recorded station.
            station = stations[station_id]

            # Deduce sensor information from the reading itself.
            for observation in reading.observations:

                # Don't list sensors twice.
                if any(map(lambda item: item.sensor_id == observation.meta.sensor_id, station['sensors'])):
                    continue

                # Build and record sensor information.
                sensor_info = Munch({
                    'sensor_id': observation.meta.sensor_id,
                    'sensor_type_name': observation.meta.sensor_type_name,
                    'sensor_type_id': observation.meta.sensor_type_id,
                })
                station['sensors'].append(sensor_info)

        # List of stations sorted by station identifier.
        results = sorted(stations.values(), key=itemgetter('station_id'))
        return results

    def get_readings_from_api(self):

        # Fetch data from remote API.
        log.info('Requesting luftdaten.info live API at {}'.format(self.uri))
        payload = self.session.get(self.uri).content.decode('utf-8')
        data = json.loads(payload)
        #pprint(data)

        # Mungle timestamp to be formally in ISO 8601 format (UTC).
        timestamp = self.convert_timestamp(data[0]['timestamp'])
        log.info('Timestamp of first record: {}'.format(timestamp))

        # Apply data filter.
        data = self.apply_filter(data)

        # Transform live API items to actual readings while optionally
        # applying a number of transformation and enrichment steps.
        for item in self.wrap_progress(data):
            try:
                reading = self.make_reading(item)
                if reading is None:
                    continue

                if not self.make_observations_from_api(item, reading):
                    continue

                log.debug(f'API reading:\n{json.dumps(reading, indent=2)}')

                yield reading

            except Exception as ex:
                log.warning('Could not make reading from {}.\n{}'.format(item, exception_traceback()))

    def make_observations_from_api(self, item, reading):

        # Collect sensor values.
        has_data = False
        if 'sensordatavalues' in item:
            for sensor in item['sensordatavalues']:
                fieldname = sensor['value_type']
                value = sensor['value']

                # Skip NaN values.
                # Prevent "influxdb.exceptions.InfluxDBClientError: 400".
                # {"error":"partial write: unable to parse 'ldi_readings,geohash=srxwdb8ny7j6,sensor_id=22674,station_id=11504 humidity=nan,temperature=18.8 1556133497000000000': invalid number
                # unable to parse 'ldi_readings,geohash=srxwdb8ny7j6,sensor_id=22674,station_id=11504 humidity=nan,temperature=18.6 1556133773000000000': invalid number dropped=0"}
                # TODO: Emit warning here?
                if is_nan(value):
                    continue

                value = float(value)

                # Actually use this reading. LDI has single observations only.
                reading.observations[0]['data'][fieldname] = value

                # Signal positive dataness.
                has_data = True

        # Skip this observation if it contains no data at all.
        if not has_data:
            log.warning('Observation without sensor values for station %s', reading.station)

        return has_data

    def filter_rule(self, data):

        for item in data:

            #log.info('item: %s', item)

            # Decode JSON item
            country_code = item['location']['country'].upper()
            station_id = item['location']['id']
            sensor_id = item['sensor']['id']

            # If there is a filter defined, evaluate it.
            # Skip further processing for specific country codes, station ids or sensor ids.
            # TODO: Improve evaluating conditions.
            skip = False
            if 'country' in self.filter:
                if country_code not in self.filter['country']:
                    skip = True
            if 'station' in self.filter:
                if station_id not in self.filter['station']:
                    skip = True
            if 'sensor' in self.filter:
                if sensor_id not in self.filter['sensor']:
                    skip = True

            if skip:
                continue

            yield item

    def make_reading(self, item):

        log.debug('Making reading from item: %s', item)

        # Decode JSON item.
        station_id = item['location']['id']
        sensor_id = item['sensor']['id']
        sensor_type_name = item['sensor']['sensor_type']['name']
        sensor_type_id = item['sensor']['sensor_type'].get('id')

        # Build observation.
        observation = Munch(
            meta=Munch(),
            data=Munch(),
        )
        entry = Munch(
            station=Munch(),
            # LDI just has single observations.
            observations=[observation],
        )

        # Set station metadata.
        entry.station.station_id = station_id

        # Set observation metadata.
        observation.meta.timestamp = self.convert_timestamp(item['timestamp'])
        observation.meta.sensor_id = sensor_id
        observation.meta.sensor_type_name = sensor_type_name
        observation.meta.sensor_type_id = sensor_type_id

        # Collect position information.
        del item['location']['id']
        entry.station.position = Munch()
        for key, value in item['location'].items():
            if key in ['latitude', 'longitude', 'altitude']:
                try:
                    value = float(value)
                except:
                    value = None
            entry.station.position[key] = value

        # Add more detailed location information.
        self.enrich_station(entry.station)

        #log.debug('Observation: %s', json.dumps(observation))

        # Debugging.
        #break

        return entry

    @staticmethod
    def convert_timestamp(timestamp):
        # Mungle timestamp to be formally in ISO 8601 format (UTC).
        if ' ' in timestamp:
            timestamp = timestamp.replace(' ', 'T')
        if '+' not in timestamp:
            timestamp += 'Z'
        return timestamp

    def get_readings_from_csv(self, path):

        # Optionally append suffix appropriately.
        suffix = '.csv'
        if not path.endswith(suffix):
            path += '/**/*' + suffix

        log.info('Building list of CSV files from {}'.format(path))
        data = list(find_files(path))
        log.info('Processing {} files'.format(len(data)))

        # Process all files, optionally add progressbar indicator.
        for csvpath in self.wrap_progress(data):
            readings = self.import_csv(csvpath)
            if readings is None:
                continue

            # Optionally, return first reading only.
            # This is a shortcut-option for churning through the whole data set quickly.
            # This can be used to get whole regional coverage of the data while lacking many details.
            if self.quick_mode:
                readings = list(readings)
                if readings:
                    yield readings[0]
                    continue

            # Process all readings per CSV file.
            for reading in readings:
                if reading is None:
                    continue
                yield reading

    @staticmethod
    def sensor_matches(path, sensors):
        for sensor in sensors:
            if sensor in path:
                return True
        return False

    def import_csv(self, csvpath):

        # Skip files not matching filter pattern (by sensor id).
        if not self.import_archive_filename_accepted(csvpath):
            return

        # When progressbar is enabled, log this to the DEBUG level.
        logger = log.info
        if self.progressbar:
            logger = log.debug
        logger('Reading CSV file {}'.format(csvpath))

        if self.sensor_matches(csvpath, ['ppd42ns', 'sds011', 'pms3003', 'pms5003', 'pms7003', 'hpm']):
            """
            PPD42NS
            sensor_id;sensor_type;location;lat;lon;timestamp;P1;durP1;ratioP1;P2;durP2;ratioP2
            27;PPD42NS;18;;;2015-10-01T19:41:50.266431+00:00;0.62;0;0;0.62;0;0

            SDS011
            sensor_id;sensor_type;location;lat;lon;timestamp;P1;durP1;ratioP1;P2;durP2;ratioP2
            92;SDS011;42;48.800;9.003;2016-07-14T13:02:46.949036+00:00;7.71;;;5.03;;

            PMS3003
            sensor_id;sensor_type;location;lat;lon;timestamp;P1;P2;P0
            409;PMS3003;193;14.382;121.047;2017-12-31T00:00:21;48.00;39.00;

            PMS5003
            sensor_id;sensor_type;location;lat;lon;timestamp;P1;P2;P0
            11390;PMS5003;5751;52.015;5.432;2018-04-20T00:00:25;45.25;31.50;22.50

            PMS7003
            sensor_id;sensor_type;location;lat;lon;timestamp;P1;P2;P0
            5920;PMS7003;2986;51.213;6.808;2017-12-27T00:01:57;5.00;4.00;1.00

            HPM
            sensor_id;sensor_type;location;lat;lon;timestamp;P1;P2
            7096;HPM;3590;50.853;4.342;2017-11-25T00:04:50;22;21
            """
            fieldnames = ['P0', 'P1', 'P2']

        elif self.sensor_matches(csvpath, ['dht22', 'htu21d']):
            """
            DHT22
            sensor_id;sensor_type;location;lat;lon;timestamp;temperature;humidity
            48;DHT22;19;48.722;9.209;2016-08-13T00:00:26.053188+00:00;;

            HTU21D
            sensor_id;sensor_type;location;lat;lon;timestamp;temperature;humidity
            2875;HTU21D;1445;52.600;13.327;2017-11-26T00:00:38;2.80;99.90
            """
            fieldnames = ['temperature', 'humidity']

        elif self.sensor_matches(csvpath, ['bmp180', 'bmp280', 'bme280']):
            """
            BMP180
            sensor_id;sensor_type;location;lat;lon;timestamp;pressure;altitude;pressure_sealevel;temperature
            1464;BMP180;725;48.410;9.934;2017-04-11T00:00:45;96011;;;11.30

            BMP280
            sensor_id;sensor_type;location;lat;lon;timestamp;pressure;altitude;pressure_sealevel;temperature
            2184;BMP280;1098;48.939;8.961;2017-11-26T00:00:35;;;;2.60

            BME280
            sensor_id;sensor_type;location;lat;lon;timestamp;pressure;altitude;pressure_sealevel;temperature;humidity
            1093;BME280;535;48.746;9.211;2017-03-31T00:00:47;96913.02;;;12.75;63.35
            """
            fieldnames = ['temperature', 'humidity', 'pressure', 'altitude', 'pressure_sealevel']

        elif self.sensor_matches(csvpath, ['ds18b20']):
            """
            DS18B20
            sensor_id;sensor_type;location;lat;lon;timestamp;temperature
            19319;DS18B20;9794;43.851;125.304;2019-01-02T00:04:02;-14.94
            """
            fieldnames = ['temperature']

        else:
            log.warning('Skip import of {}. Unknown sensor type'.format(csvpath))
            return

        return self.csv_reader(csvpath, fieldnames)

    def csv_reader(self, csvpath, fieldnames):

        payload = None
        if self.quick_mode:
            try:
                payload = open(csvpath).read(256)
                payload = '\n'.join(payload.split('\n')[:2])
            except:
                pass
        else:
            payload = open(csvpath).read()

        if payload is None:
            log.error(f'Could not read CSV file "{csvpath}')
            return

        # Read CSV file into tablib's Dataset and cast to dictionary representation.
        try:
            imported_data = Dataset().load(payload, format='csv', delimiter=';')
        except Exception:
            log.exception(f'Error decoding CSV from file {csvpath}')
            return

        records = imported_data.dict

        # Iterate all CSV records.
        for record in records:

            try:
                # Make item from CSV record.
                item = self.read_csv_record(record)

                # Make reading from item.
                reading = self.make_reading(item)
                if reading is None:
                    continue

                if not self.make_observations(record, reading, fieldnames):
                    continue

                log.debug(f'CSV reading:\n{json.dumps(reading, indent=2)}')

                yield reading

            except Exception:
                log.exception(f'Could not make observation from CSV record {record}')

    def make_observations(self, record, reading, fieldnames):

        # Indicator whether we found any data.
        has_data = False

        # Read all the CSV record fields.
        for fieldname in fieldnames:

            # Skip missing keys.
            if fieldname not in record:
                continue

            # Sanitize value.
            value = record[fieldname].strip()

            # Skip empty or non-numeric values.
            # TODO: Emit warning here?
            if is_nan(value):
                continue

            # Skip non-float values.
            # TODO: Emit warning here?
            try:
                value = float(value)
            except:
                pass

            # Actually use this reading. LDI has single observations only.
            reading.observations[0]['data'][fieldname] = value

            # Signal positive dataness.
            has_data = True

        return has_data

    def read_csv_record(self, csvitem):

        #log.debug('CSV item: %s', csvitem)

        item = {
            'timestamp': csvitem['timestamp'],
            'location': {
                'id': int(csvitem['location']),
            },
            'sensor': {
                'id': int(csvitem['sensor_id']),
                'sensor_type': {
                    'name': csvitem['sensor_type'],
                }
            },
        }

        if csvitem['lat'] and csvitem['lon']:
            item['location'].update(latitude=float(csvitem['lat']), longitude=float(csvitem['lon']))

        return item

    def import_archive_filename_accepted(self, csvpath):
        # If there is a filter defined, evaluate it.
        # For specific location|sensor ids, skip further processing.
        if self.filter:

            if 'sensor' in self.filter:
                # Decode station id from filename, e.g. ``2017-01-13_dht22_sensor_318.csv``.
                m = re.match('.+sensor_(\d+)\.csv$', csvpath)
                if m:
                    sensor_id = int(m.group(1))
                    if sensor_id not in self.filter['sensor']:
                        return False

            if 'sensor-type' in self.filter:
                # Decode sensor type from filename, e.g. ``2017-01-13_dht22_sensor_318.csv``.
                m = re.match('.+_(\w+)_sensor_\d+\.csv$', csvpath)
                if m:
                    sensor_type = m.group(1)
                    if sensor_type not in self.filter['sensor-type']:
                        return False

        return True



# -*- coding: utf-8 -*-
# (c) 2017-2019 Andreas Motl <andreas@hiveeyes.org>
# (c) 2017-2019 Richard Pobering <richard@hiveeyes.org>
# (c) 2019 Matthias Mehldau <wetter@hiveeyes.org>
# License: GNU Affero General Public License, Version 3
import logging
from copy import deepcopy
from datetime import datetime, timedelta
from operator import itemgetter

from requests import HTTPError
from rfc3339 import rfc3339
from munch import munchify, Munch
from urllib.parse import urljoin
from collections import OrderedDict

from luftdatenpumpe.source.common import AbstractLuftdatenPumpe
from luftdatenpumpe.util import slugify, grouper, chunks

log = logging.getLogger(__name__)


class IrcelinePumpe(AbstractLuftdatenPumpe):
    """
    Ingest air quality measurements from the
    Belgian Interregional Environment Agency (IRCEL - CELINE).

    - http://www.irceline.be/en
    - http://deus.irceline.be/
    - https://en.vmm.be/
    - https://project-corona.eu/

    - http://geo.irceline.be/sos/static/doc/api-doc/

    - https://wiki.52north.org/SensorWeb/SensorObservationService
    - https://wiki.52north.org/SensorWeb/SensorObservationServiceIV
    - https://wiki.52north.org/SensorWeb/SensorWebClientRESTInterfaceV0
    - https://github.com/52North/SOS

    """

    # Sensor network identifier.
    network = 'irceline'

    # IRCELINE SOS REST API URI
    uri = 'http://geo.irceline.be/sos/api/v1/'

    timeout = 60

    def get_index(self):
        return self.send_request()

    def send_request(self, endpoint=None, params=None):
        url = urljoin(self.uri, endpoint)
        log.debug(f'Requesting IRCELINE live API at {url}')
        params = params or {}

        response = self.session.get(url, params=params, timeout=self.timeout)
        if response.status_code != 200:
            try:
                reason = response.json()
            except:
                reason = 'unknown'
            message = f'Request failed: {reason}'
            log.error(message)
            response.raise_for_status()

        return response.json()

    def get_stations(self):
        """
        http://geo.irceline.be/sos/api/v1/stations/?service=1&expand=true&locale=en
        """

        data = self.send_request('stations', params={'expanded': 'true'})
        #import sys, json; print(json.dumps(data, indent=2)); sys.exit(0)

        timeseries_index = self.get_timeseries_index()
        #import sys, json; print(json.dumps(timeseries_index, indent=2)); sys.exit(0)

        # Apply data filter.
        data = self.apply_filter(data)

        stations = []
        for upstream_station in self.wrap_progress(data):
            upstream_station = munchify(upstream_station)
            #print('upstream_station:', upstream_station)
            station_info = munchify({
                'station_id': upstream_station.properties.id,
                'station_label': upstream_station.properties.label,
                'position': {
                    'country': 'BE',
                    'latitude': upstream_station.geometry.coordinates[1],
                    'longitude': upstream_station.geometry.coordinates[0],
                    'altitude': None,
                }
            })
            self.enrich_station(station_info)

            station_info.sensors = self.timeseries_to_sensors(upstream_station.properties.timeseries)

            # Enrich sensor information by timestamps of first / most recent reading.
            for sensor in station_info.sensors:
                timeseries_id = sensor['sensor_id']
                if timeseries_id in timeseries_index:
                    sensor['sensor_first_date'] = self.convert_timestamp(timeseries_index[timeseries_id].firstValue.timestamp)
                    sensor['sensor_last_date'] = self.convert_timestamp(timeseries_index[timeseries_id].lastValue.timestamp)

            #print(station_info)
            stations.append(station_info)

        # List of stations sorted by station identifier.
        stations = sorted(stations, key=itemgetter('station_id'))

        return stations

    def filter_rule(self, data):

        if self.filter and 'country' in self.filter and self.filter.country != ['BE']:
            raise NotImplementedError("Filtering by country not supported for IRCELINE/SOS, it's Belgium at all.")

        for item in data:

            #log.info('item: %s', item)

            # Decode JSON item
            #country_code = item['location']['country'].upper()
            station_id = item['properties']['id']
            sensor_ids = map(int, item['properties']['timeseries'].keys())

            # If there is a filter defined, evaluate it.
            # Skip further processing for specific country codes, station ids or sensor ids.
            # TODO: Improve evaluating conditions.
            skip = False
            #if 'country' in self.filter:
            #    if country_code not in self.filter['country']:
            #        skip = True
            if 'station' in self.filter:
                if station_id not in self.filter['station']:
                    skip = True
            if 'sensor' in self.filter:
                skip = True
                for sensor_id in sensor_ids:
                    if sensor_id in self.filter['sensor']:
                        skip = False
                        break

            if skip:
                continue

            yield item

    def timeseries_to_sensors(self, measurements):
        """

        Ingress
        ======
        ::

          "timeseries": {
              "7149": {
                "service": {
                  "id": "1",
                  "label": "IRCEL - CELINE: timeseries-api (SOS 2.0)"
                },
                "offering": {
                  "id": "7149",
                  "label": "7149 - GRIMM - procedure"
                },
                "feature": {
                  "id": "1219",
                  "label": "45R510 - Châtelineau"
                },
                "procedure": {
                  "id": "7149",
                  "label": "7149 - GRIMM - procedure"
                },
                "phenomenon": {
                  "id": "6001",
                  "label": "Particulate Matter < 2.5 µm"
                },
                "category": {
                  "id": "6001",
                  "label": "Particulate Matter < 2.5 µm"
                }
              }
          }


        Egress
        ======
        ::

            "sensors": [
                {
                    "sensor_id": 24685,
                    "sensor_type": "SDS011"
                },
                {
                    "sensor_id": 24686,
                    "sensor_type": "DHT22"
                }
            ]

        """

        # Flatten nested data a bit.
        sensors = []

        # Stable
        property_names = ['service', 'category', 'phenomenon']

        # Unstable / Legacy
        property_names_legacy = ['offering', 'feature', 'procedure']

        # TODO: Which ones to use? Use all for now.
        property_names += property_names_legacy

        for sensor_id, timeseries in measurements.items():
            sensor = OrderedDict()
            sensor['sensor_id'] = int(sensor_id)
            for property_name in property_names:
                sensor[f'sensor_{property_name}_id'] = int(timeseries[property_name]['id'])
                sensor[f'sensor_{property_name}_label'] = timeseries[property_name]['label']

            sensor['sensor_type_name'] = sensor['sensor_category_label']
            sensor['sensor_type_id'] = sensor['sensor_category_id']

            sensor['sensor_label'] = sensor['sensor_phenomenon_label']
            sensor['sensor_fieldname'] = self.slugify_fieldname(sensor['sensor_label'])

            sensors.append(sensor)

        return sensors

    def get_phenomena(self):
        """
        http://geo.irceline.be/sos/api/v1/phenomena/?service=1&locale=en

        {
            "id": "5",
            "label": "Particulate Matter < 10 \u00b5m"
        },
        {
            "id": "6001",
            "label": "Particulate Matter < 2.5 \u00b5m"
        },
        {
            "id": "6002",
            "label": "Particulate Matter < 1 \u00b5m"
        },
        """
        return self.send_request('phenomena')

    def get_timeseries_single(self):
        """
        http://geo.irceline.be/sos/api/v1/timeseries/?service=1&phenomenon=6001&expanded=true&force_latest_values=true&status_intervals=true&rendering_hints=true&locale=en
        """

    def get_readings_from_api(self):
        """
        Should yield list of dictionaries like::

          {
            "station": {
              "station_id": 1071,
              "position": {
                "latitude": 52.544,
                "longitude": 13.374,
                "altitude": 38.7,
                "country": "DE",
                "exact_location": 0,
                "geohash": "u33dbm6duz90"
              }
            },
            "observations": [
              {
                "meta": {
                  "timestamp": "2019-04-25T04:36:26Z",
                  "sensor_id": 2130,
                  "sensor_type": "DHT22"
                },
                "data": {
                  "humidity": 82.3,
                  "temperature": 17.8
                }
              }
            ]
          }

        """
        timeseries_id_list = []

        # Map stations and timeseries to their sensors.
        station_map = {}
        timeseries_sensor_map = {}
        for station in self.get_stations():
            station_id = station.station_id
            station_map[station_id] = station

            for sensor in station.sensors:
                timeseries_id = sensor['sensor_id']
                timeseries_id_list.append(timeseries_id)
                timeseries_sensor_map[timeseries_id] = sensor

        # Fetch timeseries for all stations.
        # Attention: While this is just a single call for the client,
        # this operation might be expensive at the backend.
        # TODO: We might move to chunked transfer in the future.
        #print('timeseries_id_list:', timeseries_id_list)

        # Debugging
        #timeseries = self.get_timeseries(timeseries_ids=[6895, 6931], timespan=self.filter.get('timespan'))

        # Fails
        #timeseries = self.get_timeseries(timeseries_ids=[1180, 6895], timespan=self.filter.get('timespan'))

        # For real
        timeseries = self.get_timeseries_details(timeseries_ids=timeseries_id_list, timespan=self.filter.get('timespan'))

        # Map timeseries to readings.
        timeseries_readings_map = {}
        for tsid, item in timeseries.items():
            #print(tsid, item)
            timeseries_id = int(tsid)
            fieldname = timeseries_sensor_map[timeseries_id]['sensor_fieldname']
            data = self.reading_data_from_timeseries(fieldname, item['values'])
            data = list(data)
            # TODO: Emit warning here if station has no data whatsoever?
            if not data:
                continue
            timeseries_readings_map[timeseries_id] = data

        # Make readings for all stations, grouping by timestamp.
        items = []
        for station_id in sorted(station_map.keys()):
            station = station_map[station_id]
            all_data = []
            for sensor in station.sensors:

                timeseries_id = sensor['sensor_id']
                if timeseries_id not in timeseries_readings_map:
                    log.warning(f'Station "{station_id}" has no data for timeseries "{timeseries_id}"')
                    continue

                all_data += timeseries_readings_map[timeseries_id]

            item_station = deepcopy(station)
            del item_station['sensors']

            # Group by timestamp.
            timestamp_data_map = {}
            for data_item in all_data:
                timestamp = data_item['timestamp']
                #del data_item['timestamp']
                timestamp_data_map.setdefault(timestamp, {})
                timestamp_data_map[timestamp].update(data_item)

            # Build list of observations.
            observations = []
            for data in sorted(timestamp_data_map.values(), key=itemgetter('timestamp')):
                timestamp = data['timestamp']
                del data['timestamp']
                observation = munchify(dict(meta={'timestamp': timestamp}, data=data))
                observations.append(observation)

            # List of all readings by timestamp, ascending.
            item = Munch({
                'station': item_station,
                'observations': observations,
            })
            items.append(item)

        return items

    def get_timeseries_index(self, timespan=None):
        if timespan is None:
            timespan = f'PT12h/{self.this_hour()}'

        url = urljoin(self.uri, f'timeseries/')
        data = self.send_request(url, params={'timespan': timespan, 'expanded': 'true'})
        #print(data)

        items = {}
        for entry in data:
            key = int(entry['id'])
            items[key] = munchify(entry)

        return items

    def get_timeseries_details(self, timeseries_ids=None, timespan=None):
        """
        - http://geo.irceline.be/sos/static/doc/api-doc/#timeseries-example-post-data
        - https://wiki.52north.org/SensorWeb/SensorWebClientRESTInterfaceV0#Timeseries_Data
        - https://en.wikipedia.org/wiki/ISO_8601#Time_intervals

        Valid "timespan" values are:
        - 2019-04-20T01:00:00+02:00/2019-04-22T01:00:00+02:00
        - 2019-04-20T01:00:00Z/PT0h
        - PT6h/2013-08-16TZ
        - P0Y0M3D/2013-01-31TZ

        Examples

        Data from specified station and timespan::

            http http://geo.irceline.be/sos/api/v1/timeseries/6151/getData?timespan=PT6H/2019-05-18T09:00:00Z

        - Three hours worth of data from designated starting point for a specific timeseries::

            http POST 'http://geo.irceline.be/sos/api/v1/timeseries/getData' timeseries:='[6643]' timespan='2019-04-21T22:00:00+02:00/PT3h'

        - Get readings from two timeseries at specific time::

            http POST 'http://geo.irceline.be/sos/api/v1/timeseries/getData' timeseries:='[6643,6693]' timespan='2019-04-20T01:00:00Z/PT0h'

        """

        if timespan is None:
            timespan = f'PT12h/{self.this_hour()}'

        log.info(f'Requesting IRCELINE live API with timespan "{timespan}" and {len(timeseries_ids)} timeseries')

        results = {}
        for identifier in self.wrap_progress(sorted(timeseries_ids)):
            url = urljoin(self.uri, f'timeseries/{identifier}/getData')
            log.debug(f'Requesting SOS timeseries {identifier} from {url}')
            try:
                data = self.send_request(url, params={'timespan': timespan, 'expanded': 'true'})
                results.update(data)
            except KeyboardInterrupt:
                raise
            except HTTPError as ex:
                log.error(f'Requesting data for timeseries {identifier} failed: {ex}')
            except:
                log.exception(f'Decoding response for timeseries {identifier} failed')

        return results

    def reading_data_from_timeseries(self, name, values):
        """
        Ingress
        =======
        ::

            values = [{'timestamp': 1556064000000, 'value': 172.0}, {'timestamp': 1556067600000, 'value': 162.0}]

        From::

            {
              "values": [
                {
                  "timestamp": 1556056800000,
                  "value": 36.1
                },
                {
                  "timestamp": 1556060400000,
                  "value": 34.6
                },
                {
                  "timestamp": 1556064000000,
                  "value": 33.1
                },
                {
                  "timestamp": 1556067600000,
                  "value": 32.6
                }
              ]
            }

        Egress
        ======
        ::

            {
              "time": "2019-04-24T00:52:32Z",
              "P1": 4.5,
              "P2": 3.5
            }

        """
        for observation in values:

            value = observation['value']

            # Skip empty or non-numeric values.
            # TODO: Emit warning here?
            if value is None or str(value).lower() == 'nan':
                continue

            entry = {
                'timestamp': self.convert_timestamp(observation['timestamp']),
                name: value,
            }
            yield entry

    @staticmethod
    def convert_timestamp(timestamp):
        """
        SOS timestamp are milliseconds epoch.
        """
        datetime_object = datetime.fromtimestamp(timestamp / 1000)
        return rfc3339(datetime_object)

    @staticmethod
    def this_hour():
        now = datetime.now()
        now_aligned_to_hour = now - timedelta(minutes=now.minute, seconds=now.second, microseconds=now.microsecond)
        return rfc3339(now_aligned_to_hour)

    @staticmethod
    def slugify_fieldname(fieldname):
        return slugify(fieldname)



# -*- coding: utf-8 -*-
# (c) 2018 Andreas Motl <andreas@hiveeyes.org>
# License: GNU Affero General Public License, Version 3
class StreamTarget:

    capabilities = ['stations', 'readings']

    def __init__(self, handle, formatter):
        self.handle = handle
        self.formatter = formatter
        self.buffer = []

    def emit(self, data):
        self.buffer.append(data)

    def flush(self, final=False):
        if not final:
            return
        self.handle.write(self.formatter(self.buffer))
        self.buffer = []
